/**
 * The main purpose of this module is to provide utility functions
 * that follows the restrictions of the Bookshelf/Mapper/Serializer APIs
 * with the goal of simplifying the logic of the main 'map' method.
 */
'use strict';
var lodash_1 = require('lodash');
var type_check_1 = require('type-check');
var links_1 = require('./links');
var extras_1 = require('./extras');
/**
 * Start the data processing with top level information,
 * then handle resources recursively in processResource
 */
function processData(info, data) {
    var enableLinks = info.bookOpts.enableLinks, linkOpts = info.linkOpts;
    var template = processResource(info, data);
    if (enableLinks) {
        template.dataLinks = links_1.dataLinks(linkOpts);
        template.topLevelLinks = links_1.topLinks(linkOpts);
    }
    return template;
}
exports.processData = processData;
/**
 * Recursively adds data-related properties to the
 * template to be sent to the serializer
 */
function processResource(info, data) {
    var bookOpts = info.bookOpts, linkOpts = info.linkOpts;
    var enableLinks = bookOpts.enableLinks;
    var sample = getSample(data);
    var template = {};
    // Add list of valid attributes
    template.attributes = getAttrsList(sample);
    // Nested relations (recursive) template generation
    lodash_1.forOwn(sample.relations, function (relData, relName) {
        if (!relationAllowed(bookOpts, relName)) {
            return;
        }
        var relLinkOpts = lodash_1.assign(lodash_1.clone(linkOpts), { type: relName });
        var relTemplate = processResource({ bookOpts: bookOpts, linkOpts: relLinkOpts }, relData);
        relTemplate.ref = 'id'; // Add reference in nested resources
        // Related links
        if (enableLinks) {
            relTemplate.relationshipLinks = links_1.relationshipLinks(linkOpts, relName);
            relTemplate.includedLinks = links_1.includedLinks(relLinkOpts);
        }
        template[relName] = relTemplate;
        template.attributes.push(relName);
    });
    return template;
}
exports.processResource = processResource;
/**
 * Get model sample from data to generate a template
 * Notice this method is quite type-hacky and is meant for solving null issues
 */
function getSample(data) {
    if (extras_1.isModel(data)) {
        return data;
    }
    else if (extras_1.isCollection(data)) {
        return data.reduce(lodash_1.merge, {});
    }
    else {
        return {};
    }
}
/**
 * Retrieve model's attribute names
 * following filtering rules
 */
function getAttrsList(data) {
    var attrs = lodash_1.keys(data.attributes);
    var restricted = [
        /^id$/,
        /[_-]id$/,
        /[_-]type$/
    ];
    // Only return attributes that don't match any pattern
    return attrs.filter(function (attr) {
        return !restricted.some(function (pattern) { return attr.search(pattern) >= 0; });
    });
}
/**
 * Based on Bookshelf options, determine if a relation must be included
 */
function relationAllowed(bookOpts, relName) {
    var relations = bookOpts.relations;
    return relations === true ||
        (type_check_1.typeCheck('[String]', relations) &&
            relations.some(function (rel) { return rel === relName; }));
}
/**
 * Convert a bookshelf model or collection to
 * json adding the id attribute if missing
 */
function toJSON(data) {
    var json = null;
    if (extras_1.isModel(data)) {
        json = data.serialize({ shallow: true }); // serialize without the relations
        // Assign the id for the model if it's not present already
        if (!lodash_1.has(json, 'id')) {
            json.id = data.id;
        }
        // Loop over model relations to call toJSON recursively on them
        lodash_1.forOwn(data.relations, function (relData, relName) {
            json[relName] = toJSON(relData);
        });
    }
    else if (extras_1.isCollection(data)) {
        // Run a recursive toJSON on each model of the collection
        json = data.map(toJSON);
    }
    return json;
}
exports.toJSON = toJSON;
//# sourceMappingURL=utils.js.map